package eu.jonahbauer.raytracing.scene.hittable3d;

import eu.jonahbauer.raytracing.math.AABB;
import eu.jonahbauer.raytracing.math.Range;
import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.render.material.Material;
import eu.jonahbauer.raytracing.scene.HitResult;
import eu.jonahbauer.raytracing.scene.Hittable;
import eu.jonahbauer.raytracing.scene.Target;
import eu.jonahbauer.raytracing.scene.util.PdfUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;
import java.util.Optional;
import java.util.random.RandomGenerator;

public final class Box implements Hittable, Target {
    private final @NotNull AABB box;
    private final @Nullable Material @NotNull[] materials;

    public Box(@NotNull Vec3 a, @NotNull Vec3 b, @NotNull Material material) {
        this(new AABB(a, b), material);
    }

    public Box(@NotNull AABB box, @NotNull Material material) {
        this(box, Objects.requireNonNull(material, "material"), material, material, material, material, material);
    }

    public Box(
            @NotNull Vec3 a, @NotNull Vec3 b,
            @Nullable Material top, @Nullable Material bottom,
            @Nullable Material left, @Nullable Material right,
            @Nullable Material front, @Nullable Material back
    ) {
        this(new AABB(a, b), top, bottom, left, right, front, back);
    }

    public Box(
            @NotNull AABB box,
            @Nullable Material top, @Nullable Material bottom,
            @Nullable Material left, @Nullable Material right,
            @Nullable Material front, @Nullable Material back
    ) {
        this.box = Objects.requireNonNull(box, "box");
        this.materials = new Material[] { left, bottom, back, right, top, front };
    }

    @Override
    public @NotNull Optional<HitResult> hit(@NotNull Ray ray, @NotNull Range range) {
        // based on AABB#hit with additional detection of the side hit
        var origin = ray.origin();
        var direction = ray.direction();
        var invDirection = direction.inv();

        var tmin = AABB.intersect(box.min(), origin, invDirection);
        var tmax = AABB.intersect(box.max(), origin, invDirection);

        double tlmax = Double.NEGATIVE_INFINITY;
        double tumin = Double.POSITIVE_INFINITY;

        Side entry = null;
        Side exit = null;

        for (int i = 0; i < 3; i++) {
            if (direction.get(i) >= 0) {
                if (tmin[i] > tlmax) {
                    tlmax = tmin[i];
                    entry = Side.NEGATIVE[i];
                }
                if (tmax[i] < tumin) {
                    tumin = tmax[i];
                    exit = Side.POSITIVE[i];
                }
            } else {
                if (tmax[i] > tlmax) {
                    tlmax = tmax[i];
                    entry = Side.POSITIVE[i];
                }
                if (tmin[i] < tumin) {
                    tumin = tmin[i];
                    exit = Side.NEGATIVE[i];
                }
            }
        }

        if (tlmax < tumin && tumin >= range.min() && tlmax <= range.max()) {
            assert entry != null && exit != null;
            return hit0(tlmax, tumin, entry, exit, ray, range);
        } else {
            return Optional.empty();
        }
    }

    private @NotNull Optional<HitResult> hit0(double tmin, double tmax, @NotNull Side entry, @NotNull Side exit, @NotNull Ray ray, @NotNull Range range) {
        double t;
        Side side;
        boolean frontFace;
        Material material;
        Vec3 normal;
        if (range.surrounds(tmin) && materials[entry.ordinal()] != null) {
            t = tmin;
            side = entry;
            frontFace = true;
            material = materials[side.ordinal()];
            normal = side.normal;
        } else if (range.surrounds(tmax) && materials[exit.ordinal()] != null) {
            t = tmax;
            side = exit;
            frontFace = false;
            material = materials[side.ordinal()];
            normal = side.normal.neg();
        } else {
            return Optional.empty();
        }

        var position = ray.at(t);
        var uv = material.texture().isUVRequired();
        var u = uv ? side.getTextureU(box, position) : Double.NaN;
        var v = uv ? side.getTextureV(box, position) : Double.NaN;
        return Optional.of(new HitResult(t, position, normal, this, material, u, v, frontFace));
    }

    @Override
    public @NotNull AABB getBoundingBox() {
        return box;
    }

    @Override
    public double getProbabilityDensity(@NotNull Vec3 origin, @NotNull Vec3 direction) {
        if (contains(origin)) return 1 / (4 * Math.PI);
        if (hit(new Ray(origin, direction)).isEmpty()) return 0;

        var solidAngle = 0d;
        for (var s : Side.values()) {
            if (!s.isExterior(box, origin)) continue;
            solidAngle += s.getSolidAngle(box, origin);
        }

        return 1 / solidAngle;
    }

    @Override
    public @NotNull Vec3 getTargetingDirection(@NotNull Vec3 origin, @NotNull RandomGenerator random) {
        if (contains(origin)) return Vec3.random(random);

        // determine sides facing the origin and their solid angles
        int visible = 0;

        // at most three faces are visible
        Side[] sides = new Side[3];
        double[] solidAngle = new double[3];

        double accumSolidAngle = 0;
        for (var s : Side.values()) {
            if (!s.isExterior(box, origin)) continue;

            var sa = s.getSolidAngle(box, origin);
            accumSolidAngle += sa;
            sides[visible] = s;
            solidAngle[visible] = accumSolidAngle;
            visible++;
        }

        // choose a random side facing the origin based on their relative solid angles
        var r = random.nextDouble() * solidAngle[visible - 1];
        for (int j = 0; j < visible; j++) {
            if (r < solidAngle[j]) {
                // choose a random point on that side
                var target = sides[j].random(box, random);
                return target.minus(origin);
            }
        }

        throw new AssertionError();
    }

    private boolean contains(@NotNull Vec3 point) {
        return box.min().x() < point.x() && point.x() < box.max().x()
               && box.min().y() < point.y() && point.y() < box.max().y()
               && box.min().z() < point.z() && point.z() < box.max().z();
    }

    @Override
    public @NotNull String toString() {
        return "Box(min=" + box.min() + ", max=" + box.max() + ")";
    }

    private enum Side {
        NEG_X(Vec3.UNIT_X.neg()),
        NEG_Y(Vec3.UNIT_Y.neg()),
        NEG_Z(Vec3.UNIT_Z.neg()),
        POS_X(Vec3.UNIT_X),
        POS_Y(Vec3.UNIT_Y),
        POS_Z(Vec3.UNIT_Z),
        ;

        private static final Side[] NEGATIVE = new Side[] {Side.NEG_X, Side.NEG_Y, Side.NEG_Z};
        private static final Side[] POSITIVE = new Side[] {Side.POS_X, Side.POS_Y, Side.POS_Z};

        private final @NotNull Vec3 normal;

        Side(@NotNull Vec3 normal) {
            this.normal = Objects.requireNonNull(normal, "normal");
        }

        /**
         * {@return the texture u coordinate for a position on this side of the box}
         */
        public double getTextureU(@NotNull AABB box, @NotNull Vec3 pos) {
            return switch (this) {
                case NEG_X -> (pos.z() - box.min().z()) / (box.max().z() - box.min().z());
                case POS_X -> (box.max().z() - pos.z()) / (box.max().z() - box.min().z());
                case NEG_Y, POS_Y, POS_Z -> (pos.x() - box.min().x()) / (box.max().x() - box.min().x());
                case NEG_Z -> (box.max().x() - pos.x()) / (box.max().x() - box.min().x());
            };
        }

        /**
         * {@return the texture v coordinate for a position on this side of the box}
         */
        public double getTextureV(@NotNull AABB box, @NotNull Vec3 pos) {
            return switch (this) {
                case NEG_X, POS_X, NEG_Z, POS_Z -> (pos.y() - box.min().y()) / (box.max().y() - box.min().y());
                case NEG_Y -> (pos.z() - box.min().z()) / (box.max().z() - box.min().z());
                case POS_Y -> (box.max().z() - pos.z()) / (box.max().z() - box.min().z());
            };
        }

        /**
         * {@return whether the given position is outside of the box only considering <code>this</code> side}
         */
        public boolean isExterior(@NotNull AABB box, @NotNull Vec3 pos) {
            return switch (this) {
                case NEG_X -> pos.x() < box.min().x();
                case NEG_Y -> pos.y() < box.min().y();
                case NEG_Z -> pos.z() < box.min().z();
                case POS_X -> pos.x() > box.max().x();
                case POS_Y -> pos.y() > box.max().y();
                case POS_Z -> pos.z() > box.max().z();
            };
        }

        /**
         * {@return the point on <code>this</code> side of the <code>box</code> with the given <code>u</code>-<code>v</code>-coordinates}
         */
        public @NotNull Vec3 get(@NotNull AABB box, double u, double v) {
            return switch (this) {
                case NEG_X -> new Vec3(
                        box.min().x(),
                        Math.fma(v, box.max().y() - box.min().y(), box.min().y()),
                        Math.fma(u, box.max().z() - box.min().z(), box.min().z())
                );
                case NEG_Y -> new Vec3(
                        Math.fma(u, box.max().x() - box.min().x(), box.min().x()),
                        box.min().y(),
                        Math.fma(v, box.max().z() - box.min().z(), box.min().z())
                );
                case NEG_Z -> new Vec3(
                        Math.fma(u, box.min().x() - box.max().x(), box.max().x()),
                        Math.fma(v, box.max().y() - box.min().y(), box.min().y()),
                        box.min().z()
                );
                case POS_X -> new Vec3(
                        box.max().x(),
                        Math.fma(v, box.max().y() - box.min().y(), box.min().y()),
                        Math.fma(u, box.min().z() - box.max().z(), box.max().z())
                );
                case POS_Y -> new Vec3(
                        Math.fma(u, box.max().x() - box.min().x(), box.min().x()),
                        box.max().y(),
                        Math.fma(v, box.min().z() - box.max().z(), box.max().z())
                );
                case POS_Z -> new Vec3(
                        Math.fma(u, box.max().x() - box.min().x(), box.min().x()),
                        Math.fma(v, box.max().y() - box.min().y(), box.min().y()),
                        box.max().z()
                );
            };
        }

        /**
         * {@return a random point on <code>this</code> side of the <code>box</code>}
         */
        public @NotNull Vec3 random(@NotNull AABB box, @NotNull RandomGenerator random) {
            var u = random.nextDouble();
            var v = random.nextDouble();
            return get(box, u, v);
        }

        /**
         * {@return the solid angle covered by <code>this</code> side of the <code>box</code> when viewed from <code>pos</code>}
         */
        public double getSolidAngle(@NotNull AABB box, @NotNull Vec3 pos) {
            var a = get(box, 0, 0);
            var b = get(box, 0, 1);
            var c = get(box, 1, 1);
            var d = get(box, 1, 0);

            return PdfUtil.getSolidAngle(pos, a, b, d) + PdfUtil.getSolidAngle(pos, c, b, d);
        }
    }
}
